--------------------------------------------------------------------------------
-- CERN (BE-CO-HT)
-- WB slave test bench for vme64x core
-- http://www.ohwr.org/projects/svec
--------------------------------------------------------------------------------
--
-- unit name:     vmespy
--
-- author:        Tristan Gingold <tristan.gingold@cern.ch>
--
-- description:
--
--   WB slave to be synthetized to test features of the vme64x core.
--
-- dependencies:
--
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;

entity vmespy is

  port(
    clk_sys_i : in std_logic;
    rst_n_i   : in std_logic;

    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;

    int_o   : out  std_logic;

    leds_o : out std_logic_vector(15 downto 0);

    vme_write_n_i    : in    std_logic;
    vme_lword_n_i    : in    std_logic;
    vme_iack_n_i     : in    std_logic;
    vme_dtack_n_i    : in    std_logic;
    vme_ds_n_i       : in    std_logic_vector(1 downto 0);
    vme_as_n_i       : in    std_logic;
    vme_data_i       : in    std_logic_vector(31 downto 0);
    vme_am_i         : in    std_logic_vector(5 downto 0);
    vme_addr_i       : in    std_logic_vector(31 downto 1)
);

end vmespy;

architecture rtl of vmespy is
  --  Memory map (WB byte addresses):
  --  0 - 0x3ff: sram (512*4B)
  --   0x4000: leds (4B)
  --   0x4004: last WB transaction (see the code for the format)
  --   0x4008: nbr of WB read accesses (write to clear)
  --   0x400c: nbr of WB write accesses (likewise)
  --   0x4010: nbr of write errors in pattern ram
  --   0x4014: generates bus error.
  --   0x4018: spy status
  --   0x401c: spy counters
  --   0x8000: counter (4B). Generate an interrupt when 0 is reached.
  --   0xc000: pattern ram (0x1000 * 4B)
  --  0x20000: spy ram
  --  0x40000 - 0x3ff000: pattern ram
  signal counter : unsigned(31 downto 0);
  signal leds : std_logic_vector(15 downto 0);

  signal last_trans : std_logic_vector (28 downto 0);

  signal nbr_read : unsigned (15 downto 0);
  signal nbr_write: unsigned (15 downto 0);
  signal nbr_write_errors: unsigned (31 downto 0);

  signal pattern : std_logic_vector (31 downto 0);

  type sram_type is array (0 to 16#1ff#) of std_logic_vector(31 downto 0);
  signal sram: sram_type;


  --  Post synchronizer signals.
  signal vme_write_n    : std_logic;
  signal vme_lword_n    : std_logic;
  signal vme_iack_n     : std_logic;
  signal vme_dtack_n     : std_logic;
  signal vme_ds_n       : std_logic_vector(1 downto 0);
  signal vme_as_n       : std_logic;
  signal vme_data       : std_logic_vector(31 downto 0);
  signal vme_am         : std_logic_vector(5 downto 0);
  signal vme_addr       : std_logic_vector(31 downto 1);

  signal vme_as_n_d     : std_logic;

  --  spy state
  --  Spy is enabled, waiting for trigger.  Cleared when triggered.
  signal spy_wait : std_logic;
  --  Spy has been triggered, writing to RAM.
  signal spy_write : std_logic;
  --  Spy done.
  signal spy_done : std_logic;
  --  Number of transfers (AS rising edge before writing to RAM)
  signal spy_count : unsigned(7 downto 0);
  constant c_spy_depth : natural := 13;
  --  Current RAM index (while writing).
  signal spy_index : unsigned(c_spy_depth - 1 downto 0);

  signal spy_start : std_logic;
  signal spy_write_reg : std_logic;
  signal spy_ram_data : std_logic_vector (75 downto 0);
begin

  --  Pattern of the pattern ram.
  pattern (31 downto 16) <= not slave_i.adr(15 downto 0);
  pattern (15 downto 0)  <= slave_i.adr(15 downto 0);

  process (clk_sys_i)
    procedure pattern_write
    is
      variable err : boolean;
    begin
      err := false;
      for i in 3 downto 0 loop
        if slave_i.sel (i) = '1'
          and (slave_i.dat(8*i + 7 downto 8*i) /=
               pattern (8*i + 7 downto 8*i))
        then
          err := true;
        end if;
      end loop;
      if err then
        nbr_write_errors <= nbr_write_errors + 1;
      end if;
    end pattern_write;

    variable idx : natural;

    variable to_ack : std_logic_vector (1 downto 0);
  begin
    if rising_edge(clk_sys_i) then
      slave_o.ack <= '0';
      slave_o.err <= '0';

      if rst_n_i = '0' then
        counter <= (others => '0');
        leds <= (others => '0');
        nbr_read <= (others => '0');
        nbr_write <= (others => '0');
        nbr_write_errors <= (others => '0');
        spy_start <= '0';
        spy_write_reg <= '0';
        to_ack := (others => '0');
      else
        -- Decrementer
        if counter /= (counter'range => '0') then
          counter <= counter - 1;
        end if;

        spy_start <= '0';
        spy_write_reg <= '0';

        if slave_i.stb = '1' and slave_i.cyc = '1' then
          if slave_i.adr (13 downto 12) = "00" then
            -- Save transaction (very cheap scope).
            last_trans (23 downto 0) <= slave_i.adr (23 downto 0);
            last_trans (27 downto 24) <= slave_i.sel;
            last_trans (28) <= slave_i.we;
          end if;

          if slave_i.we = '1' then
            --  Write
            nbr_write <= nbr_write + 1;
            if slave_i.adr(27 downto 16) = x"000" then
              case slave_i.adr (15 downto 14) is
                when "00" =>
                  idx := to_integer(unsigned(slave_i.adr(10 downto 2)));
                  for i in 3 downto 0 loop
                    if slave_i.sel (i) = '1' then
                      sram(idx)(8*i + 7 downto 8*i) <=
                        slave_i.dat(8*i + 7 downto 8*i);
                    end if;
                  end loop;
                when "01" =>
                  case slave_i.adr (4 downto 2) is
                    when "000" =>
                      --  leds
                      for i in 1 downto 0 loop
                        if slave_i.sel (i) = '1' then
                          leds(8*i + 7 downto 8*i) <=
                            slave_i.dat(8*i + 7 downto 8*i);
                        end if;
                      end loop;
                    when "001" =>
                      null;
                    when "010" =>
                      nbr_read <= (others => '0');
                    when "011" =>
                      nbr_write <= (others => '0');
                    when "100" =>
                      nbr_write_errors <= (others => '0');
                    when "101" =>
                      slave_o.err <= '1';
                    when "110" =>
                      spy_start <= slave_i.dat (0);
                    when "111" =>
                      spy_write_reg <= '1';
                    when others =>
                      null;
                  end case;
                when "10" =>
                  for i in 3 downto 0 loop
                    if slave_i.sel (i) = '1' then
                      counter(8*i + 7 downto 8*i) <=
                        unsigned(slave_i.dat(8*i + 7 downto 8*i));
                    end if;
                  end loop;
                when "11" =>
                  pattern_write;
                when others =>
                  null;
              end case;
            end if;
            to_ack := "01";
          else
            -- Read
            nbr_read <= nbr_read + 1;
            if slave_i.adr(27 downto 16) = x"000" then
              case slave_i.adr (15 downto 14) is
                when "00" =>
                  idx := to_integer(unsigned(slave_i.adr(10 downto 2)));
                  slave_o.dat <= sram(idx);
                when "01" =>
                  case slave_i.adr (4 downto 2) is
                    when "000" =>
                      slave_o.dat(31 downto 16) <= (others => '0');
                      slave_o.dat(15 downto 0) <= leds;
                    when "001" =>
                      slave_o.dat <= (31 downto 29 => '0') & last_trans;
                    when "010" =>
                      slave_o.dat (31 downto 16) <= (others => '0');
                      slave_o.dat (15 downto 0) <= std_logic_vector (nbr_read);
                    when "011" =>
                      slave_o.dat (31 downto 16) <= (others => '0');
                      slave_o.dat (15 downto 0) <= std_logic_vector (nbr_write);
                    when "100" =>
                      slave_o.dat <= std_logic_vector (nbr_write_errors);
                    when "101" =>
                      slave_o.err <= '1';
                    when "110" =>
                      slave_o.dat <= (others => '0');
                      slave_o.dat (0) <= spy_wait;
                      slave_o.dat (1) <= spy_write;
                      slave_o.dat (2) <= spy_done;
                    when "111" =>
                      slave_o.dat <= (others => '0');
                      slave_o.dat (7 downto 0) <= std_logic_vector (spy_count);
                      slave_o.dat (16 + c_spy_depth - 1 downto 16) <= std_logic_vector (spy_index);
                    when others =>
                      null;
                  end case;
                when "10" =>
                  slave_o.dat <= std_logic_vector(counter);
                when "11" =>
                  slave_o.dat <= pattern;
                when others =>
                  null;
              end case;
              to_ack := "01";
            elsif slave_i.adr(27 downto 16) = x"002"
              or slave_i.adr(27 downto 16) = x"003"
            then
              slave_o.dat <= (others => '0');
              case slave_i.adr (3 downto 2) is
                when "00" =>
                  slave_o.dat <= spy_ram_data (31 downto 0);
                when "01" =>
                  slave_o.dat <= spy_ram_data (63 downto 32);
                when "10" =>
                  slave_o.dat (11 downto 0) <= spy_ram_data (75 downto 64);
                when others =>
                  null;
              end case;
              if to_ack = "00" then
                to_ack := "10";
              end if;
            else
              slave_o.dat <= pattern;
              to_ack := "01";
            end if;
          end if;
        end if;
      end if;
      slave_o.ack <= to_ack (0);
      to_ack := '0' & to_ack (1);
    end if;
  end process;

  leds_o <= leds;
  int_o <= '1' when counter = 1 else '0';

  -- drive unused WB slave_o outputs
  slave_o.stall <= '0';
  slave_o.rty   <= '0';

  inst_vme_write_resync: entity work.gc_sync_register
    generic map (g_width => 1)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i(0) => vme_write_n_i,
              q_o(0) => vme_write_n);
  inst_vme_lword_resync: entity work.gc_sync_register
    generic map (g_width => 1)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i(0) => vme_lword_n_i,
              q_o(0) => vme_lword_n);
  inst_vme_iack_resync: entity work.gc_sync_register
    generic map (g_width => 1)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i(0) => vme_iack_n_i,
              q_o(0) => vme_iack_n);
  inst_vme_dtack_resync: entity work.gc_sync_register
    generic map (g_width => 1)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i(0) => vme_dtack_n_i,
              q_o(0) => vme_dtack_n);
  inst_vme_ds_resync: entity work.gc_sync_register
    generic map (g_width => 2)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i => vme_ds_n_i,
              q_o => vme_ds_n);
  inst_vme_as_resync: entity work.gc_sync_register
    generic map (g_width => 1)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i(0) => vme_as_n_i,
              q_o(0) => vme_as_n);
  inst_vme_data_resync: entity work.gc_sync_register
    generic map (g_width => 32)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i => vme_data_i,
              q_o => vme_data);
  inst_vme_am_resync: entity work.gc_sync_register
    generic map (g_width => 6)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i => vme_am_i,
              q_o => vme_am);
  inst_vme_addr_resync: entity work.gc_sync_register
    generic map (g_width => 31)
    port map (clk_i => clk_sys_i,
              rst_n_a_i => rst_n_i,
              d_i => vme_addr_i,
              q_o => vme_addr);

  process (clk_sys_i)
    type vram_type is array (0 to 2**c_spy_depth - 1) of std_logic_vector(75 downto 0);
    variable vram : vram_type;
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        spy_wait <= '0';
        spy_write <= '0';
        spy_done <= '0';
        spy_count <= (others => '0');
        spy_index <= (others => '0');

        vme_as_n_d <= '1';
      else
        --  DATA ram read.
        spy_ram_data <= vram (to_integer (unsigned (slave_i.adr (c_spy_depth + 3 downto 4))));

        if spy_write = '1' or spy_wait = '1' then
          --  Write when writing (!) but also the current input while waiting.
          vram (to_integer(spy_index)) :=
            (not vme_dtack_n) & (not vme_as_n) & (not vme_iack_n) & (not vme_write_n)
            & (not vme_ds_n) & vme_am
            & vme_addr & (not vme_lword_n) & vme_data;
        end if;

        if spy_write = '1' then
          --  Update memory address.
          if spy_index = (spy_index'range => '1') then
            spy_write <= '0';
            spy_done <= '1';
          else
            spy_index <= spy_index + 1;
          end if;
        end if;

        if spy_wait = '1' then
          --  Trigger ?
          if vme_as_n_d = '1' and vme_as_n = '0' then
            if spy_count = (spy_count'range => '0') then
              spy_wait <= '0';
              spy_write <= '1';
              spy_index <= to_unsigned(1, c_spy_depth);
            else
              spy_index <= (others => '0');
              spy_count <= spy_count - 1;
            end if;
          end if;
        end if;

        --  Write to registers.
        if spy_wait = '0' and spy_write = '0' then
          if spy_write_reg = '1' then
            spy_count <= unsigned (slave_i.dat (7 downto 0));
          end if;
          spy_wait <= spy_start;
          spy_done <= spy_done and not spy_start;
        end if;

        --  Prev as.
        vme_as_n_d <= vme_as_n;
      end if;
    end if;
  end process;
end rtl;
