/*
 * Copyright (C) 2017 CERN (www.cern.ch)
 * Author: Tristan Gingold <tristan.gingold@cern.ch>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * test_vme: program to test the vme64xcore (to be used with the
 * svec_vmecore_test design).
 */

#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdarg.h>
#include "libvmebus.h"

#define MAX(A, B) ((A) > (B) ? (A) : (B))

static unsigned slot = 0;
static volatile unsigned char *conf;

static int nbr_errors;

static uint32_t
read_conf4 (unsigned int addr)
{
  return (conf[addr + 0] << 24)
    | (conf[addr + 4] << 16)
    | (conf[addr + 8] << 8)
    | (conf[addr + 12] << 0);
}

static uint64_t
read_conf8 (unsigned int addr)
{
  uint64_t res = 0;
  for (int i = 0; i < 8; i++)
    res = (res << 8) | conf[addr + i * 4];
  return res;
}

struct vme_function
{
  uint64_t amcap;
  uint32_t adem;
  uint32_t ader;
  uint8_t am;
};

static void
read_vme_function (struct vme_function *res, int i)
{
  res->amcap = read_conf8(0x123 + i * 0x20);
  res->adem = read_conf4(0x623 + i * 0x10);
  res->ader = read_conf4(0x7ff63 + i * 0x10);
  res->am = (res->ader >> 2) & 0x3f;
}

static bool
is_vme_function_enabled (struct vme_function *f)
{
  if ((f->ader & 1) || ((f->amcap >> f->am) & 1) == 0)
    return false;
  else
    return true;
}

static struct vme_mapping data_map;
static void *ptr;
static volatile unsigned char *ptr8;
static volatile unsigned short *ptr16;
static volatile unsigned int *ptr32;

/* Setup the board for access mode AM.  Return the base + mask.  */

static void
setup_am (uint8_t am, uint32_t *base, uint32_t *mask)
{
  for (int i = 0; i < 8; i++)
    {
      struct vme_function fn;
      read_vme_function (&fn, i);

      if (!((fn.amcap >> am) & 1))
	continue;

      *mask = fn.adem & 0xffffff00;
      *base = (slot << 19) & *mask;
      fn.ader = *base | (am << 2);
      fn.am = am;
      conf[0x7ff63 + i * 0x10 +  0] = (fn.ader >> 24) & 0xff;
      conf[0x7ff63 + i * 0x10 +  4] = (fn.ader >> 16) & 0xff;
      conf[0x7ff63 + i * 0x10 +  8] = (fn.ader >>  8) & 0xff;
      conf[0x7ff63 + i * 0x10 + 12] = (fn.ader >>  0) & 0xff;
      return;
    }
  fprintf (stderr, "am 0x%02x is not supported\n", am);
  exit (3);
}

/* Map vme space defined by AM, BASE, MASK to user memory.  */

static void
setup_map (uint8_t am, uint32_t base, uint32_t mask)
{
  if (ptr != NULL)
    {
      vme_unmap (&data_map, 1);
      ptr = NULL;
    }

  data_map.am = am;
  data_map.data_width = 32;
  data_map.sizel = MAX (0x80000, ~mask + 1);
  data_map.vme_addrl = base & mask;

  printf ("INFO: Map VME 0x%08x AM 0x%02x\n", data_map.vme_addrl, data_map.am);

  /* Do mmap */
  ptr = vme_map(&data_map, 1);
  if (!ptr)
    {
      fprintf (stderr, "cannot map vme space\n");
      exit (3);
    }

  ptr8 = (volatile unsigned char *)ptr;
  ptr16 = (volatile unsigned short *)ptr;
  ptr32 = (volatile unsigned int *)ptr;
}

/* Setup the board for AM and map the address space.  */

static uint32_t
map_for_am (uint8_t am)
{
  uint32_t base;
  uint32_t mask;

  setup_am (am, &base, &mask);
  setup_map (am, base, mask);
  return base;
}


static unsigned int
read_nbr_interrupts (unsigned int level)
{
  const char *filename = "/proc/vme/interrupts";
  FILE *f;

  f = fopen (filename, "r");
  if (f == NULL)
    {
      fprintf (stderr, "cannot open %s\n", filename);
      exit (3);
    }
  while (1)
    {
      unsigned int l, num;
      char line[80];

      if (fgets (line, sizeof (line), f) == NULL)
	break;
      //printf ("DBG: %s", line);
      if (sscanf (line, "IRQ%u %u\n", &l, &num) == 2
	  && l == level)
	{
	  fclose (f);
	  return num;
	}
    }
  fclose (f);
  fprintf (stderr, "failed to find IRQ %u in %s\n", level, filename);
  exit (3);
}

static void
report (bool ok, const char *msg, ...)
{
  va_list args;

  va_start(args, msg);

  if (ok)
    printf ("OK:  ");
  else
    {
      printf ("ERR: ");
      nbr_errors++;
    }
  vprintf (msg, args);
  putchar ('\n');
  va_end(args);
}

static void
do_test_mem (uint8_t am)
{
#define M8(addr)  (*(volatile uint8_t  *)(ptr + addr))
#define M16(addr) (*(volatile uint16_t *)(ptr + addr))
#define M32(addr) (*(volatile uint32_t *)(ptr + addr))

  printf ("INFO: test memory access for am 0x%02x\n", am);
  map_for_am (am);

  // Clear
  for (int i = 0; i < 16; i += 4)
    M32(i) = 0;
  // Check
  for (int i = 0; i < 16; i += 4)
    if (M32(i) != 0)
      report (false, "clear at 0x%04x", i * 4);

  // Write a byte and check
  M8(3) = 0xb3;
  report (M32(0) == 0xb3000000, "write byte at 3");

  // Write a byte and check
  M8(0) = 0x27;
  report (M32(0) == 0xb3000027, "write byte at 0");

  // Write a word
  M16(4) = 0x3c8f;
  report (M32(4) == 0x3c8f, "write word at 4");
  M16(6) = 0x744d;
  report (M32(4) == 0x744d3c8f, "write word at 6");

  // Write a long word
  M32(8) = 0xedcba987;
  report (M32(8) == 0xedcba987, "write long word at 8");

  // Read byte
  report (M8(8) == 0x87, "read byte at 8");
  report (M8(11) == 0xed, "read byte at 11");
  report (M8(9) == 0xa9, "read byte at 9");
  report (M8(10) == 0xcb, "read byte at 10");

  // Read word
  report (M16(8) == 0xa987, "read word at 8");
  report (M16(10) == 0xedcb, "read word at 10");

  // Check again
  report (M32(4) == 0x744d3c8f, "write long word at 4");
  report (M32(0) == 0xb3000027, "write long byte at 0");

#undef M8
#undef M16
#undef M32
}

static void
rate_start (struct timespec *start_ts)
{
  if (clock_gettime (CLOCK_MONOTONIC, start_ts) != 0)
    {
      fprintf (stderr, "clock_gettime error: %m\n");
      exit (4);
    }
}

static void
rate_done (struct timespec *start_ts, unsigned int nbr_bytes)
{
  struct timespec end_ts;
  unsigned long nano;

  if (clock_gettime (CLOCK_MONOTONIC, &end_ts) != 0)
    {
      fprintf (stderr, "clock_gettime error: %m\n");
      exit (4);
    }
  nano = (end_ts.tv_nsec - start_ts->tv_nsec)
    + (end_ts.tv_sec - start_ts->tv_sec) * 1000000000;
  printf ("Rate: %f MB/sec\n",
	  nbr_bytes * 1.0E9 / nano / (1 << 20));
}

static void
do_test_error_counter (void)
{
  // Clear pattern write error counter
  map_for_am (0x39);
  ptr32[0x1004] = 0;
  report (ptr32[0x1004] == 0, "write error counter correctly reset");
  ptr32[0x3000] = 0x12345678;
  report (swapbe32 (ptr32[0x1004]) == 1,
	  "write error counter is working (on error)");
  ptr32[0x3000] = swapbe32 (0xcfff3000);
  report (swapbe32 (ptr32[0x1004]) == 1,
	  "write error counter is working (on correct)");
  ptr32[0x1004] = 0;
}

static void
do_test_dma_with_len (uint8_t am, uint32_t off, uint32_t len)
{
  struct vme_dma dma;
  uint32_t base;
  uint32_t mask;
  void *buf;
  int s;
  struct timespec start_ts;
  enum vme_dma_block_size bsize;
  int nbr_err;

  // Clear pattern write error counter
  map_for_am (0x39);
  ptr32[0x1004] = 0;
  if (ptr32[0x1004] != 0)
    {
      printf ("FATAL: write error counter correctly reset\n");
      exit (3);
    }

  printf ("INFO: test DMA for am 0x%02x (off: 0x%08x, len: 0x%0x)\n",
          am, off, len);
  setup_am (am, &base, &mask);

  if (posix_memalign (&buf, 4096, len) != 0)
    {
      fprintf (stderr, "cannot allocate memory\n");
      exit (4);
    }
  memset (&dma, 0, sizeof (dma));
  dma.status = 0; // ??
  dma.length = len;
  dma.novmeinc = 0;
  dma.dir = VME_DMA_FROM_DEVICE;

  dma.src.data_width = VME_D32;
  dma.src.am = am;
  dma.src.v2esst_mode = VME_SST160;
  dma.src.bcast_select = 0;
  dma.src.addru = 0;
  dma.src.addrl = base + off;

  dma.dst.addru = ((uintptr_t)buf) >> 32;
  dma.dst.addrl = (uintptr_t)buf;

  if (am == VME_A24_SUP_MBLT
      || am == VME_A24_USER_MBLT
      || am == VME_A32_SUP_MBLT
      || am == VME_A32_USER_MBLT)
    bsize = VME_DMA_BSIZE_2048;
  else
    bsize = VME_DMA_BSIZE_256;

  dma.ctrl.vme_block_size = bsize;
  dma.ctrl.vme_backoff_time = VME_DMA_BACKOFF_1;
  dma.ctrl.pci_block_size = bsize;
  dma.ctrl.pci_backoff_time = VME_DMA_BACKOFF_1;

  rate_start (&start_ts);
  s = vme_dma_read (&dma);
  rate_done (&start_ts, len);

  report (s == 0, "DMA read");

  nbr_err = 0;
  for (int j = 0; j < len / 4; j++)
    {
      uint32_t r = ((uint32_t *)buf)[j];
      uint32_t v;

      v = ((j << 2) + off) & 0xffff;
      v = v | (~v << 16);

      r = swapbe32 (r);
      if (r != v)
	{
	  report (false,
		  "dma read error at 0x%08x (got 0x%08x, expect 0x%08x)",
		  off + (j << 2), r, v);
          nbr_err++;
          if (nbr_err >= 10)
            break;
	}
    }

  //  DMA write
  memset (&dma, 0, sizeof (dma));
  dma.status = 0; // ??
  dma.length = len;
  dma.novmeinc = 0;
  dma.dir = VME_DMA_TO_DEVICE;

  dma.src.addru = ((uintptr_t)buf) >> 32;
  dma.src.addrl = (uintptr_t)buf;

  dma.dst.data_width = VME_D32;
  dma.dst.am = am;
  dma.dst.v2esst_mode = VME_SST160;
  dma.dst.bcast_select = 0;
  dma.dst.addru = 0;
  dma.dst.addrl = base + off;

  dma.ctrl.vme_block_size = bsize;
  dma.ctrl.vme_backoff_time = VME_DMA_BACKOFF_1;
  dma.ctrl.pci_block_size = bsize;
  dma.ctrl.pci_backoff_time = VME_DMA_BACKOFF_1;

  rate_start (&start_ts);
  s = vme_dma_write (&dma);
  rate_done (&start_ts, len);
  report (s == 0, "DMA write");

  free (buf);

  map_for_am (0x39);
  report (ptr32[0x1004] == 0, "no write error during DMA (%d errors)",
	  swapbe32 (ptr32[0x1004]));
}

static void
do_test_dma (uint8_t am)
{
  do_test_dma_with_len (am, 0xc000, 0x4000);
}

static void
do_irq_test (void)
{
  unsigned int level = conf[0x7ff5b];
  unsigned int num1, num2;

  // Check the level is not used.
  num1 = read_nbr_interrupts (level);
  usleep (500000);
  num2 = read_nbr_interrupts (level);
  report (num1 == num2, "IRQ line is unused");
  if (num1 != num2)
    return;

  map_for_am (0x39);
  ptr32[0x8000 / 4] = swapbe32 (20);
  usleep (1000);
  num1 = read_nbr_interrupts (level);
  report (num1 == num2 + 1, "IRQ reported");
}

static void
do_berr_test (void)
{
#define BERR_BIT 0x08
#define BIT_SET_REG 0x7fffb
#define BIT_CLR_REG 0x7fff7
  conf[BIT_CLR_REG] = BERR_BIT; // BERR
  report ((conf[BIT_SET_REG] & BERR_BIT) == 0, "BERR flag cleared (1)");
  report ((conf[BIT_CLR_REG] & BERR_BIT) == 0, "BERR flag cleared (2)");

  map_for_am (0x39);
  ptr32[0x1005] = 1;

  report ((conf[BIT_SET_REG] & BERR_BIT) != 0,
	  "BERR flag set after write error");

  conf[BIT_CLR_REG] = BERR_BIT;
  report ((conf[BIT_SET_REG] & BERR_BIT) == 0, "BERR flag cleared (3)");

  (void)ptr32[0x1005];

  report ((conf[BIT_SET_REG] & BERR_BIT) != 0,
	  "BERR flag set after read error");
  conf[BIT_CLR_REG] = BERR_BIT;
}

static int
do_vme_test (void)
{
  // A24
  // do_test_mem (0x3e);
  do_test_mem (0x39);

  // A32
  do_test_mem (0x0a);

  // DMA
  do_test_error_counter ();
  printf ("\n"
	  "PERF: SCT rate\n");
  do_test_dma (0x3d); // Single
  printf ("\n"
	  "PERF: BLT rate\n");
  do_test_dma (0x3f); // BLT
  printf ("\n"
	  "PERF: MBLT rate\n");
  do_test_dma (0x3c); // MBLT

  printf ("\n");
  do_irq_test ();
  do_berr_test ();

  if (nbr_errors == 0)
    {
      printf ("Success!\n");
      return 0;
    }
  else
    {
      printf ("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
      printf ("FAILED!!! FAILED!!! FAILED!!! FAILED!!! FAILED!!! FAILED!!!\n");
      printf ("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
      return 1;
    }
}

static void
set_leds (unsigned short v)
{
  /* 0: off,  1: green,  2: red,  3: orange.  */
  volatile unsigned short *leds = &ptr16[0x4002 >> 1];
  *leds = v;
}

static void
play_leds (const unsigned short *pat, size_t l)
{
  size_t i;
  for (i = 0; i < l; i++)
    {
      set_leds (pat[i]);
      usleep (100000);
    }
}

static void
do_leds (void)
{
  static const unsigned short pat1[] = { 0x0001, 0x0004, 0x0010, 0x0040,
					 0x0100, 0x0400, 0x1000, 0x4000,
					 0x0100, 0x0040, 0x0010 };
  static const unsigned short pat2[] = { 0x0202, 0x0808, 0x2020, 0x8080,
					 0x2020, 0x0808 };
  static const unsigned short pat3[] = { 0x7c00, 0x1e00,
					 0x0702, 0x0107, 0x002d, 0x00b4,
					 0x80d0, 0xe040};
  int j;

  for (j = 0; j < 4; j++)
    play_leds (pat1, sizeof (pat1) / sizeof (unsigned short));
  for (j = 0; j < 6; j++)
    play_leds (pat2, sizeof (pat2) / sizeof (unsigned short));
  for (j = 0; j < 8; j++)
    play_leds (pat3, sizeof (pat3) / sizeof (unsigned short));
}

void
do_show_functions (void)
{
  int i;

  for (i = 0; i < 8; i++)
    {
      int j;
      struct vme_function fn;

      read_vme_function (&fn, i);
      printf ("F%d: amcap=%016lx adem=%08x ader=%08x am=%02x\n",
              i, fn.amcap, fn.adem, fn.ader, fn.am);
      for (j = 0; j < 64; j++)
        if ((fn.amcap >> j) & 1)
          printf (" %02x", j);
      printf ("\n");
    }
}

int
main (int argc, char **argv)
{
  int am = -1;
  int c;
  struct vme_mapping conf_map;
  char *action = NULL;

  while ((c = getopt (argc, argv, "s:ha:m:")) != -1)
    switch (c)
      {
      case 's':
	slot = strtol (optarg, NULL, 0);
	break;
      case 'm':
	am = strtol (optarg, NULL, 0);
	break;
      case 'a':
	action = optarg;
	break;
      case 'h':
	printf ("usage: %s [-h] [-m AM] [-a ACTION] -s SLOT\n", argv[0]);
	return 0;
      case '?':
	fprintf (stderr, "incorrect option, try %s -h\n", argv[0]);
	return 1;
      }

  if (slot == 0 || slot > 0x1f)
    {
      fprintf (stderr, "incorrect or missing slot value (1 - 0x1f)\n");
      return 2;
    }

  /* Map the CR/CSR space of the board.  */
  memset (&conf_map, 0, sizeof (conf_map));
  conf_map.am = VME_CR_CSR;
  conf_map.data_width = 32;
  conf_map.sizel = 512 << 10;
  conf_map.vme_addrl = (slot << 19);

  conf = vme_map(&conf_map, 1);
  if (!conf)
    {
      fprintf (stderr, "cannot map vme CR/CSR space\n");
      return 3;
    }

  if (action == NULL)
    {
      printf ("No action, testing vme...\n");
      return do_vme_test();
    }
  else
    printf ("action: %s\n", action);

  /* Set an am if needed.  */
  if (am != -1)
    map_for_am (am);
  else
    {
      /* Find a map.  */
      memset (&data_map, 0, sizeof (data_map));

      for (int i = 0; i < 8; i++)
        {
          struct vme_function fn;

          read_vme_function (&fn, i);

          if (!is_vme_function_enabled (&fn))
            continue;

          uint32_t mask = fn.adem & 0xffffff00;
          uint32_t base = fn.ader & mask;

          printf ("Func %d: 0x%08x-0x%08x, am=0x%02x\n",
                  i, base, (base & mask) | ~mask, fn.am);
          setup_map (fn.am, base, mask);
          break;
        }

      if (data_map.data_width == 0)
        {
          printf ("No map found\n");
          return 3;
        }
    }

  if (action == NULL || strcmp (action, "simple") == 0)
    {
      int i;

      ptr32[0] = 0x11223344;
      ptr8[0] = 0x95;
      ptr16[0x04 >> 1] = 0x2794;
      ptr32[0x0c >> 2] = 0x13412693;

      ptr32[8 >> 2] = 0x11223344;
      ptr8[8] = 0x95;

      for (i = 0; i < 4; i++)
	printf ("%02x: %08x\n", i * 4, ptr32[i]);
    }
  else if (strcmp (action, "check-sel") == 0)
    {
      printf ("lword write\n");
      ptr32[1] = 0x12345678;
      printf ("last_trans: 0x%08x\n", swapbe32 (ptr32[0x4004 >> 2]));
      printf ("word write\n");
      ptr16[5] = 0x0123;
      printf ("last_trans: 0x%08x\n", swapbe32 (ptr32[0x4004 >> 2]));
      printf ("byte write\n");
      ptr8[13] = 0xab;
      printf ("last_trans: 0x%08x\n", swapbe32 (ptr32[0x4004 >> 2]));
    }
  else if (strcmp (action, "last-trans") == 0)
    {
      printf ("last_trans: 0x%08x\n", swapbe32 (ptr32[0x4004 >> 2]));
    }
  else if (strcmp (action, "dump") == 0)
    {
      int i;
      for (i = 0; i < 0x10; i += 4)
	printf ("%04x: %08x\n", i, ptr32[i >> 2]);
    }
  else if (strcmp (action, "rdump") == 0)
    {
      int i;
      for (i = 0x0c; i >= 0; i -= 4)
	printf ("%04x: %08x\n", i, ptr32[i >> 2]);
    }
  else if (strcmp (action, "dump0") == 0)
    {
      int i;
      for (i = 0; i < 100; i++)
	printf ("%04x: %08x\n", 0, ptr32[0]);
    }
  else if (strcmp (action, "clear") == 0)
    {
      int i;
      for (i = 0x0; i < 0x4; i++)
	ptr32[i] = 0;
    }
  else if (strcmp (action, "clear0") == 0)
    {
      int i;
      for (i = 0x0; i < 0x4; i++)
	ptr32[0] = 0;
    }
  else if (strcmp (action, "write0") == 0)
    {
      ptr32[0] = 0x10111213;
      ptr8[0] = 0x57;
    }
  else if (strcmp (action, "write800") == 0)
    {
      ptr32[0x800 >> 2] = 0x11223344;
      ptr8[0x800] = 0x95;
    }
  else if (strcmp (action, "write01") == 0)
    {
      ptr8[0] = 0x76;
      ptr32[1] = 0xabcdef98;
    }
  else if (strcmp (action, "leds") == 0)
    do_leds ();
  else if (strcmp (action, "counter") == 0)
    {
      volatile unsigned int *cntr = &ptr32[0x8000 >> 2];
      int i;

      if (*cntr == 0)
	*cntr = swapbe32 (0x01ffffff);

      for (i = 0; i < 100; i++)
	printf ("%08x\n", swapbe32 (*cntr));
    }
  else if (strcmp (action, "stats") == 0)
    {
      volatile unsigned int *stats = &ptr32[0x4000 >> 2];

      printf ("last_trans: 0x%08x\n", swapbe32 (stats[1]));
      printf ("nbr_read:   0x%08x\n", swapbe32 (stats[2]));
      printf ("nbr_write:  0x%08x\n", swapbe32 (stats[3]));
    }
  else if (strcmp (action, "clr-stats") == 0)
    {
      volatile unsigned int *stats = &ptr32[0x4000 >> 2];

      stats[2] = 0;
      stats[3] = 0;
    }
  else if (strcmp (action, "rcheck") == 0)
    {
      unsigned int i, j;
      unsigned int cnt = 4 * 16;
      struct timespec start_ts;

      rate_start (&start_ts);
      for (j = 0; j < cnt; j++)
	for (i = 0xc000; i <= 0xffff; i += 4)
	  {
	    unsigned int r = ptr32[i >> 2];

	    /* Check the read value is correct. */
	    if (0)
	      {
		unsigned int val = (i >> 2) | ((~(i >> 2)) << 16);
		r = swapbe32 (r);
		if (val != r)
		  {
		    printf ("read error at %04x: read %08x, expect %08x\n",
			    i, r, val);
		    break;
		  }
	      }
	  }
      rate_done (&start_ts, cnt * 0x4000);
    }
  else if (strcmp (action, "wcheck") == 0)
    {
      unsigned int i, j;
      unsigned int cnt = 4 * 16;
      struct timespec start_ts;

      rate_start (&start_ts);
      for (j = 0; j < cnt; j++)
	for (i = 0xc000; i <= 0xffff; i += 4)
	  ptr32[i >> 2] = i;
      rate_done (&start_ts, cnt * 0x4000);
    }
  else if (strcmp (action, "irq") == 0)
    {
      unsigned int level = conf[0x7ff5b];
      printf ("irq level:  %d\n", level);
      printf ("irq vector: %d\n", conf[0x7ff5f]);
      printf ("nbr ints: %u\n", read_nbr_interrupts (level));
    }
  else if (strcmp (action, "blt") == 0)
    {
      if (am == -1)
        am = 0x39;
      do_test_dma_with_len (am, 0x20000, 0x40000);
    }
  else if (strcmp (action, "mblt") == 0)
    {
      if (am == -1)
        am = 0x38;
      do_test_dma_with_len (am, 0x20000, 0x40000);
    }
  else if (strcmp (action, "show-func") == 0)
    do_show_functions ();
  else
    printf ("unknown action\n");

  return 0;
}
