#!/usr/bin/env python
import PyUAL
import getopt, sys
import lsvme

nbr_err = 0

def report(msg, ok):
    global nbr_err
    if ok:
        s = "OK:  "
    else:
        s = "ERR: "
        nbr_err += 1
    print(s + msg)

def do_test_cr(cr_csr):
    chksum_computed = lsvme.compute_checksum(cr_csr)
    chksum_read = cr_csr.readb(0x03)
    report("checksum", chksum_read == chksum_computed)

def do_test_mem(addr, am):
    print("INFO: memory test using mode 0x{0:02x} {1}".format(
            am, lsvme.AM_map[am]))
    vme_desc = PyUAL.PyUALVME(4, am, 0x10000, addr, 0)
    vme_acc = PyUAL.PyUAL("vme", vme_desc)

    # clear
    for i in range(0,16,4):
        vme_acc.writel(i, 0)
    # and check
    for i in range(0,16,4):
        if vme_acc.readl(i) != 0:
            report("clear", False)
    # Write a byte and check
    vme_acc.writeb(3, 0xb3)
    report("byte write at 3", vme_acc.readl(0) == 0xb3000000)
    # Write a byte and check
    vme_acc.writeb(0, 0x27)
    report("byte write at 0", vme_acc.readl(0) == 0xb3000027)
    # Write a word
    vme_acc.writew(4, 0x3c8f)
    report("word write at 4", vme_acc.readl(4) == 0x3c8f)
    # Write a word
    vme_acc.writew(6, 0x744d)
    report("word write at 6", vme_acc.readl(4) == 0x744d3c8f)
    # Write a long word
    vme_acc.writel(8, 0xedcba987)
    report("lword write at 8", vme_acc.readl(8) == 0xedcba987)
    # Read byte
    report("byte read at 8", vme_acc.readb(8) == 0x87)
    report("byte read at 11", vme_acc.readb(11) == 0xed)
    report("byte read at 9", vme_acc.readb(9) == 0xa9)
    report("byte read at 10", vme_acc.readb(10) == 0xcb)

def swap32(x):
    return (((x << 24) & 0xFF000000) |
            ((x <<  8) & 0x00FF0000) |
            ((x >>  8) & 0x0000FF00) |
            ((x >> 24) & 0x000000FF))

def do_test_blt(addr, am):
    print("INFO: BLT test using mode 0x{0:02x} {1}".format(
            am, lsvme.AM_map[am]))
    vme_desc = PyUAL.PyUALVME(4, am, 0x10000, addr, 0)
    vme_acc = PyUAL.PyUAL("vme", vme_desc)

    # clear
    for i in range(0,16,4):
        vme_acc.writel(i, 0)
    # and check
    for i in range(0,16,4):
        if vme_acc.readl(i) != 0:
            report("clear", False)
    for i in range(0xc000, 0xc010, 4):
        v = swap32(vme_acc.readl(i))
        e = ((i >> 2) | ((~(i >> 2)) << 16)) & 0xffffffff
        report("word at {0:04x} (v:{1:08x}, e:{2:08x})".format(i, v, e),
               v == e)

def main():
    slot = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "s:")
    except getopt.GetoptError as err:
        print str(err)
        sys.exit(2)
    for o, a in opts:
        if o == "-s":
            slot = int(a)
        else:
            assert False, "unhandled option"
    
    if not slot:
        print ("missing -s SLOT option")
        sys.exit(1)

    cr_csr_desc = PyUAL.PyUALVME(4, 0x2f, 0x80000, slot << 19, 0)
    cr_csr = PyUAL.PyUAL("vme", cr_csr_desc)

    if not lsvme.validate_cr(cr_csr):
        print ("CR not validated")
        sys.exit(1)

    do_test_cr(cr_csr)

    ader = lsvme.read_ader(cr_csr, 1)
    adem = lsvme.read_adem(cr_csr, 1)
    am = ader['am']
    if am != 0x39:
        print("unexpected AM")
    addr = adem['addr'] & ader['addr']
    print("INFO: addr: 0x{0:08x}, am: 0x{1:02x} {2}".format(
            addr, am, lsvme.AM_map[am]))

    am = 0x39
    cr_csr.writeb(0x7ff63 + 1 * 0x10 + 0xc, am << 2)
    do_test_mem(addr, am)

    am = 0x3e
    cr_csr.writeb(0x7ff63 + 1 * 0x10 + 0xc, am << 2)
    do_test_mem(addr, am)

    am = 0x3f
    cr_csr.writeb(0x7ff63 + 1 * 0x10 + 0xc, am << 2)
    do_test_blt(addr, am)

    # Restore AM
    am = 0x39
    cr_csr.writeb(0x7ff63 + 1 * 0x10 + 0xc, am << 2)

    # Report
    if nbr_err == 0:
        print("OK!!")
        sys.exit(0)
    else:
        print("FAILED!!")
        sys.exit(1)

if __name__ == "__main__":
    main()

