/*
 * Copyright (C) 2017 CERN (www.cern.ch)
 * Author: Tristan Gingold <tristan.gingold@cern.ch>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * test_vme: program to test the vme64xcore (to be used with the
 * svec_vmecore_test design).
 */

#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdarg.h>
#include "libvmebus.h"

#define MAX(A, B) ((A) > (B) ? (A) : (B))

static unsigned slot = 0;
static volatile unsigned char *conf;

static uint32_t
read_conf4 (unsigned int addr)
{
  return (conf[addr + 0] << 24)
    | (conf[addr + 4] << 16)
    | (conf[addr + 8] << 8)
    | (conf[addr + 12] << 0);
}

static uint64_t
read_conf8 (unsigned int addr)
{
  uint64_t res = 0;
  for (int i = 0; i < 8; i++)
    res = (res << 8) | conf[addr + i * 4];
  return res;
}

struct vme_function
{
  uint64_t amcap;
  uint32_t adem;
  uint32_t ader;
  uint8_t am;
};

static void
read_vme_function (struct vme_function *res, int i)
{
  res->amcap = read_conf8(0x123 + i * 0x20);
  res->adem = read_conf4(0x623 + i * 0x10);
  res->ader = read_conf4(0x7ff63 + i * 0x10);
  res->am = (res->ader >> 2) & 0x3f;
}

static bool
is_vme_function_enabled (struct vme_function *f)
{
  if ((f->ader & 1) || ((f->amcap >> f->am) & 1) == 0)
    return false;
  else
    return true;
}

static struct vme_mapping data_map;
static void *ptr;
static volatile unsigned char *ptr8;
static volatile unsigned short *ptr16;
static volatile unsigned int *ptr32;

/* Setup the board for access mode AM.  Return the base + mask.  */

static void
setup_am (uint8_t am, uint32_t *base, uint32_t *mask)
{
  for (int i = 0; i < 8; i++)
    {
      struct vme_function fn;
      read_vme_function (&fn, i);

      if (!((fn.amcap >> am) & 1))
	continue;

      *mask = fn.adem & 0xffffff00;
      *base = (slot << 19) & *mask;
      fn.ader = *base | (am << 2);
      fn.am = am;
      conf[0x7ff63 + i * 0x10 +  0] = (fn.ader >> 24) & 0xff;
      conf[0x7ff63 + i * 0x10 +  4] = (fn.ader >> 16) & 0xff;
      conf[0x7ff63 + i * 0x10 +  8] = (fn.ader >>  8) & 0xff;
      conf[0x7ff63 + i * 0x10 + 12] = (fn.ader >>  0) & 0xff;
      return;
    }
  fprintf (stderr, "am 0x%02x is not supported\n", am);
  exit (3);
}

/* Map vme space defined by AM, BASE, MASK to user memory.  */

static void
setup_map (uint8_t am, uint32_t base, uint32_t mask)
{
  if (ptr != NULL)
    {
      vme_unmap (&data_map, 1);
      ptr = NULL;
    }

  data_map.am = am;
  data_map.data_width = 32;
  data_map.sizel = MAX (0x80000, ~mask + 1);
  data_map.vme_addrl = base & mask;

  printf ("INFO: Map VME 0x%08x AM 0x%02x (base=%08x mask=%08x)\n",
          data_map.vme_addrl, data_map.am, base, mask);

  /* Do mmap */
  ptr = vme_map(&data_map, 1);
  if (!ptr)
    {
      fprintf (stderr, "cannot map vme space\n");
      exit (3);
    }

  ptr8 = (volatile unsigned char *)ptr;
  ptr16 = (volatile unsigned short *)ptr;
  ptr32 = (volatile unsigned int *)ptr;
}

/* Setup the board for AM and map the address space.  */

static uint32_t
map_for_am (uint8_t am)
{
  uint32_t base;
  uint32_t mask;

  setup_am (am, &base, &mask);
  setup_map (am, base, mask);
  return base;
}

static void
map_crcsr (void)
{
  struct vme_mapping conf_map;

  /* Map the CR/CSR space of the board.  */
  memset (&conf_map, 0, sizeof (conf_map));
  conf_map.am = VME_CR_CSR;
  conf_map.data_width = 32;
  conf_map.sizel = 512 << 10;
  conf_map.vme_addrl = (slot << 19);

  conf = vme_map(&conf_map, 1);
  if (!conf)
    {
      fprintf (stderr, "cannot map vme CR/CSR space\n");
      exit (3);
    }
}

void
do_show_functions (void)
{
  int i;

  for (i = 0; i < 8; i++)
    {
      int j;
      struct vme_function fn;

      read_vme_function (&fn, i);
      printf ("F%d: amcap=%016lx adem=%08x ader=%08x am=%02x\n",
              i, fn.amcap, fn.adem, fn.ader, fn.am);
      for (j = 0; j < 64; j++)
        if ((fn.amcap >> j) & 1)
          printf (" %02x", j);
      printf ("\n");
    }
}

int
main (int argc, char **argv)
{
  int am = -1;
  int c;
  char *action = NULL;

  while ((c = getopt (argc, argv, "s:ha:m:")) != -1)
    switch (c)
      {
      case 's':
	slot = strtol (optarg, NULL, 0);
	break;
      case 'm':
	am = strtol (optarg, NULL, 0);
	break;
      case 'a':
	action = optarg;
	break;
      case 'h':
	printf ("usage: %s [-h] [-m AM] [-a ACTION] -s SLOT\n", argv[0]);
	return 0;
      case '?':
	fprintf (stderr, "incorrect option, try %s -h\n", argv[0]);
	return 1;
      }

  if (slot == 0 || slot > 0x1f)
    {
      fprintf (stderr, "incorrect or missing slot value (1 - 0x1f)\n");
      return 2;
    }

  if (action == NULL)
    action = "regs";

  if (strcmp (action, "set-ader") == 0)
    {
      unsigned int fn;
      uint32_t addr;
      unsigned int am;

      if (optind + 3 != argc)
        {
          fprintf (stderr, "usage: set-adr FN ADDR AM\n");
          exit (2);
        }
      fn = strtoul (argv[optind + 0], NULL, 0);
      addr = strtoul (argv[optind + 1], NULL, 0);
      am = strtoul (argv[optind + 2], NULL, 0);

      printf ("setting ader %u to %08x %02x\n", fn, addr, am);

      map_crcsr ();

      addr |= am << 2;

      conf[0x7ff63 + fn * 0x10 +  0] = (addr >> 24) & 0xff;
      conf[0x7ff63 + fn * 0x10 +  4] = (addr >> 16) & 0xff;
      conf[0x7ff63 + fn * 0x10 +  8] = (addr >>  8) & 0xff;
      conf[0x7ff63 + fn * 0x10 + 12] = (addr >>  0) & 0xff;

      return 0;
    }
  else if (strcmp (action, "show-ader") == 0)
    {
      map_crcsr ();
      do_show_functions ();
      return 0;
    }

  if (am == -1)
    {
      /* MAP A24.  */
      setup_map (0x39, slot << 19, ~((512 << 10) - 1));
    }
  else
    {
      map_crcsr ();

      /* Set an am if needed.  */
      if (am != -1)
        map_for_am (am);
      else
        {
          /* Find a map.  */
          memset (&data_map, 0, sizeof (data_map));

          for (int i = 0; i < 8; i++)
            {
              struct vme_function fn;

              read_vme_function (&fn, i);

              if (!is_vme_function_enabled (&fn))
                continue;

              uint32_t mask = fn.adem & 0xffffff00;
              uint32_t base = fn.ader & mask;

              printf ("Func %d: 0x%08x-0x%08x, am=0x%02x\n",
                      i, base, (base & mask) | ~mask, fn.am);
              setup_map (fn.am, base, mask);
              break;
            }

          if (data_map.data_width == 0)
            {
              printf ("No map found\n");
              return 3;
            }
        }
    }

  if (action == NULL || strcmp (action, "regs") == 0)
    {
      printf ("spy status:   0x%08x\n", swapbe32 (ptr32[0x4018 >> 2]));
      printf ("spy counters: 0x%08x\n", swapbe32 (ptr32[0x401c >> 2]));
    }
  else if (strcmp (action, "count") == 0)
    {
      ptr32[0x401c >> 2] = swapbe32 (20);
    }
  else if (strcmp (action, "start") == 0)
    {
      unsigned int count = 10;

      if (optind < argc)
        count = atoi (argv[optind]);

      printf ("Start trigger after %u transfers\n", count);
      ptr32[0x401c >> 2] = swapbe32 (count);
      ptr32[0x4018 >> 2] = swapbe32 (1);
    }
  else if (strcmp (action, "dump") == 0)
    {
      unsigned last;
      unsigned addr, prev_addr;
      unsigned data, prev_data;
      unsigned ctrl, prev_ctrl;

      last = swapbe32 (ptr32[0x401c >> 2]) >> 16;
      prev_ctrl = 0;
      prev_data = 0;
      prev_addr = 0;
      printf ("cycl:  data     addr     am as wr ds dtack\n");
      for (unsigned i = 0; i <= last; i++)
        {
          data = swapbe32 (ptr32[(0x20000 + (i << 4) + 0) >> 2]);
          addr = swapbe32 (ptr32[(0x20000 + (i << 4) + 4) >> 2]);
          ctrl = swapbe32 (ptr32[(0x20000 + (i << 4) + 8) >> 2]);
          if (ctrl != prev_ctrl || addr != prev_addr || data != prev_data)
            {
              printf ("%04u: ", i);
              printf (" %08x", data);
              printf (" %08x", addr ^ 1);
              printf (" %02x  %x  %x  %x %x\n",
                      ctrl & 0x3f, (ctrl >> 10) & 1, (ctrl >> 8) & 1,
                      (ctrl >> 6) & 3, (ctrl >> 11) & 1);

              prev_ctrl = ctrl;
              prev_addr = addr;
              prev_data = data;
            }
        }
    }
  else
    printf ("unknown action\n");

  return 0;
}
